import os
from subprocess import CompletedProcess
import subprocess
import sys
from typing import Union, Literal, Optional, Dict
import json


def python_run(
    cmdline,
    returncode: Union[Literal["error"], Literal["any"], int],
    cwd=None,
    additional_env: Optional[Dict[str, str]] = None,
    timeout=None,
) -> CompletedProcess:
    cp = os.environ.copy()
    cp["PYTHONPATH"] = os.pathsep.join([x for x in sys.path if x])
    if additional_env:
        cp.update(additional_env)
    args = [sys.executable] + cmdline
    result = subprocess.run(args, capture_output=True, env=cp, cwd=cwd, timeout=timeout)

    if returncode == "any":
        return result

    if returncode == "error" and result.returncode:
        return result

    if result.returncode == returncode:
        return result

    # This is a bit too verbose, so, commented out for now.
    # env_str = "\n".join(str(x) for x in sorted(cp.items()))

    raise AssertionError(
        f"""Expected returncode: {returncode}. Found: {result.returncode}.
=== stdout:
{result.stdout.decode('utf-8')}

=== stderr:
{result.stderr.decode('utf-8')}

=== Args:
{args}

"""
    )


project_rootdir = os.path.abspath(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
cwd = os.path.join(project_rootdir, "tests_runfiles", "samples_integrated")
runfiles = os.path.join(project_rootdir, "runfiles.py")
test_in_root = os.path.join(cwd, "root", "test_in_root.py")

assert os.path.exists(cwd), f"{cwd} does not exist."
assert os.path.exists(runfiles), f"{runfiles} does not exist."
assert os.path.exists(test_in_root), f"{test_in_root} does not exist."
assert os.path.exists(project_rootdir), f"{project_rootdir} does not exist."


env_filter = {
    "PYDEV_RUNFILES_FILTER_TESTS": json.dumps(
        {
            "include": [
                [
                    test_in_root,
                    "MyTest.test1",
                ],
            ]
        }
    )
}


def test_runfiles_integrated():
    use_env = {}
    use_env.update(env_filter)
    use_env["PYTHONPATH"] = cwd
    completed = python_run(["-Xfrozen_modules=off", runfiles, cwd], returncode=0, cwd=cwd, additional_env=use_env)
    # print(completed.stdout.decode("utf-8"))
    # print(completed.stderr.decode("utf-8"))
    output = completed.stdout.decode("utf-8")
    assert "--on-test-1" in output
    assert "--on-test-2" not in output
    assert "--on-test-3" not in output



import os

import _debugger_case_sysexit_unhandled_break

# Raise an exception in a system module.
def raise_exception():
    # This code runs in debugpy when attaching. This mimics the behavior of debugpy
    # so we can test that exceptions are ignored properly.
    importlib_metadata = None
    try:
        import importlib_metadata
    except ImportError:  # pragma: no cover
        try:
            from importlib import metadata as importlib_metadata
        except ImportError:
            pass
    if importlib_metadata is None:  # pragma: no cover
        print("Cannot enumerate installed packages - missing importlib_metadata.")
    else:
        print("Installed packages:\n")
        try:
            for pkg in importlib_metadata.distributions():
                print("    {0}=={1}\n", pkg.name, pkg.version)
        except Exception:  # pragma: no cover
            print(
                "Error while enumerating installed packages."
            )
raise_exception()

current_path = os.path.dirname(os.path.abspath(__file__))
runner_path = os.path.join(current_path, '_debugger_case_sysexit_unhandled_attach.py')

# Use pydevd to run the other module. This is how debugpy runs pydevd
import _pydevd_bundle.pydevd_runpy
_pydevd_bundle.pydevd_runpy.run_path(runner_path) # final break




r"""
Full setup, used to distribute the debugger backend to PyPi.

Note that this is mostly so that users can do:

pip install pydevd

in a machine for doing remote-debugging, as a local installation with the IDE should have
everything already distributed.

Reference on wheels:
https://hynek.me/articles/sharing-your-labor-of-love-pypi-quick-and-dirty/
http://lucumr.pocoo.org/2014/1/27/python-on-wheels/

Another (no wheels): https://jamie.curle.io/blog/my-first-experience-adding-package-pypi/

See:

build_tools\pydevd_release_process.txt

for release process.
"""

from setuptools import setup
from setuptools.dist import Distribution
from distutils.extension import Extension
import os


class BinaryDistribution(Distribution):
    def is_pure(self):
        return False


data_files = []

this_dir = os.path.dirname(os.path.abspath(__file__))
attach_to_process_dir = os.path.join(this_dir, "pydevd_attach_to_process")


def accept_file(f):
    f = f.lower()
    for ext in ".py .dll .so .dylib .txt .cpp .h .bat .c .sh .md .txt".split():
        if f.endswith(ext):
            return True

    return f in ["readme", "makefile"]


def make_rel(p):
    return os.path.relpath(p, this_dir)


data_files.append(
    (
        "pydevd_attach_to_process",
        [make_rel(os.path.join(attach_to_process_dir, f)) for f in os.listdir(attach_to_process_dir) if accept_file(f)],
    )
)
for root, dirs, files in os.walk(attach_to_process_dir):
    for d in dirs:
        data_files.append(
            (
                make_rel(os.path.join(root, d)),
                [make_rel(os.path.join(root, d, f)) for f in os.listdir(os.path.join(root, d)) if accept_file(f)],
            )
        )

import pydevd

version = pydevd.__version__

with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "README.md"), "r", encoding="utf-8") as stream:
    long_description = stream.read()

args = dict(
    name="pydevd",
    version=version,
    description="PyDev.Debugger (used in PyDev, PyCharm and VSCode Python)",
    long_description_content_type="text/markdown",
    long_description=long_description,
    author="Fabio Zadrozny and others",
    url="https://github.com/fabioz/PyDev.Debugger/",
    license="EPL",
    packages=[
        "_pydev_bundle",
        "_pydev_bundle.fsnotify",
        "_pydev_runfiles",
        "_pydevd_bundle",
        "_pydevd_bundle._debug_adapter",
        "_pydevd_bundle.pydevd_concurrency_analyser",
        "_pydevd_frame_eval",
        "_pydevd_sys_monitoring",
        "_pydevd_frame_eval.vendored",
        "_pydevd_frame_eval.vendored.bytecode",
        "pydev_ipython",
        # 'pydev_sitecustomize', -- Not actually a package (not added)
        "pydevd_attach_to_process",
        "pydevd_plugins",
        "pydevd_plugins.extensions",
        "pydevd_plugins.extensions.types",
    ],
    py_modules=[
        # 'interpreterInfo', -- Not needed for debugger
        # 'pycompletionserver', -- Not needed for debugger
        "pydev_app_engine_debug_startup",
        # 'pydev_coverage', -- Not needed for debugger
        # 'pydev_pysrc', -- Not needed for debugger
        "pydev_run_in_console",
        "pydevconsole",
        "pydevd_file_utils",
        "pydevd",
        "pydevd_tracing",
        # 'runfiles', -- Not needed for debugger
        "setup_pydevd_cython",  # Distributed to clients. See: https://github.com/fabioz/PyDev.Debugger/issues/102
        # 'setup', -- Should not be included as a module
    ],
    classifiers=[
        "Development Status :: 6 - Mature",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Eclipse Public License 1.0 (EPL-1.0)",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Software Development :: Debuggers",
    ],
    entry_points={
        "console_scripts": [
            "pydevd = pydevd:main",
        ],
    },
    data_files=data_files,
    keywords=["pydev", "pydevd", "pydev.debugger"],
    include_package_data=True,
    zip_safe=False,
)

import sys

try:
    extra_compile_args = []
    extra_link_args = []

    if "linux" in sys.platform:
        # Enabling -flto brings executable from 4MB to 0.56MB and -Os to 0.41MB
        # Profiling shows an execution around 3-5% slower with -Os vs -O3,
        # so, kept only -flto.
        extra_compile_args = ["-flto", "-O3"]
        extra_link_args = extra_compile_args[:]

        # Note: also experimented with profile-guided optimization. The executable
        # size became a bit smaller (from 0.56MB to 0.5MB) but this would add an
        # extra step to run the debugger to obtain the optimizations
        # so, skipped it for now (note: the actual benchmarks time was in the
        # margin of a 0-1% improvement, which is probably not worth it for
        # speed increments).
        # extra_compile_args = ["-flto", "-fprofile-generate"]
        # ... Run benchmarks ...
        # extra_compile_args = ["-flto", "-fprofile-use", "-fprofile-correction"]
    elif "win32" in sys.platform:
        pass
        # uncomment to generate pdbs for visual studio.
        # extra_compile_args=["-Zi", "/Od"]
        # extra_link_args=["-debug"]

    kwargs = {}
    if extra_link_args:
        kwargs["extra_link_args"] = extra_link_args
    if extra_compile_args:
        kwargs["extra_compile_args"] = extra_compile_args

    ext_modules = [
        # In this setup, don't even try to compile with cython, just go with the .c file which should've
        # been properly generated from a tested version.
        Extension(
            "_pydevd_bundle.pydevd_cython",
            [
                "_pydevd_bundle/pydevd_cython.c",
            ],
            define_macros=[("Py_BUILD_CORE_MODULE", "1")],
            **kwargs,
        )
    ]

    py_version = sys.version_info[:2]
    if (3, 6) <= py_version <= (3, 10):
        ext_modules.append(
            Extension(
                "_pydevd_frame_eval.pydevd_frame_evaluator",
                [
                    "_pydevd_frame_eval/pydevd_frame_evaluator.c",
                ],
                define_macros=[("Py_BUILD_CORE_MODULE", "1")],
                **kwargs,
            )
        )

    # Note: 3.11 does not have frame eval implemented (nor sys.monitoring)

    if py_version >= (3, 12):
        ext_modules.append(
            Extension(
                "_pydevd_sys_monitoring._pydevd_sys_monitoring_cython",
                [
                    "_pydevd_sys_monitoring/_pydevd_sys_monitoring_cython.c",
                ],
                define_macros=[("Py_BUILD_CORE_MODULE", "1")],
                **kwargs,
            )
        )

    args_with_binaries = args.copy()
    args_with_binaries.update(dict(distclass=BinaryDistribution, ext_modules=ext_modules))
    setup(**args_with_binaries)
except:
    # Compile failed: just setup without compiling cython deps.
    setup(**args)
    sys.stdout.write("Plain-python version of pydevd installed (cython speedups not available).\n")



# coding: utf-8
import sys
from _pydevd_bundle.pydevd_constants import EXCEPTION_TYPE_USER_UNHANDLED
import pytest
from tests_python.debug_constants import IS_PY311_OR_GREATER


def test_create_frames_list_from_traceback():
    def method():
        raise RuntimeError("first")

    def method1():
        try:
            method()
        except Exception as e:
            raise RuntimeError("second") from e

    def method2():
        try:
            method1()
        except Exception as e:
            raise RuntimeError("third") from e

    try:
        method2()
    except Exception as e:
        exc_type, exc_desc, trace_obj = sys.exc_info()
        frame = sys._getframe()

        from _pydevd_bundle.pydevd_frame_utils import create_frames_list_from_traceback

        frames_list = create_frames_list_from_traceback(trace_obj, frame, exc_type, exc_desc, exception_type=EXCEPTION_TYPE_USER_UNHANDLED)
        assert str(frames_list.exc_desc) == "third"
        assert str(frames_list.chained_frames_list.exc_desc) == "second"
        assert str(frames_list.chained_frames_list.chained_frames_list.exc_desc) == "first"
        assert frames_list.chained_frames_list.chained_frames_list.chained_frames_list is None


if IS_PY311_OR_GREATER:
    import traceback

    _byte_offset_to_character_offset = getattr(traceback, "_byte_offset_to_character_offset", None)
    if _byte_offset_to_character_offset is not None:
        _original = traceback._byte_offset_to_character_offset

        def _byte_offset_to_character_offset(*args, **kwargs):
            try:
                return _original(*args, **kwargs)
            except:
                # Replacement to deal with the buggy version released on Python 3.11.0.
                def replacement(str, offset):
                    as_utf8 = str.encode("utf-8")
                    if offset > len(as_utf8):
                        offset = len(as_utf8)

                    return len(as_utf8[: offset + 1].decode("utf-8", "replace"))

                return replacement(*args, **kwargs)

        traceback._byte_offset_to_character_offset = _byte_offset_to_character_offset

_USE_UNICODE = [False, True]


@pytest.mark.parametrize("use_unicode", _USE_UNICODE)
@pytest.mark.skipif(not IS_PY311_OR_GREATER, reason="Python 3.11 required.")
def test_collect_anchors_subscript(use_unicode):
    from _pydevd_bundle.pydevd_frame_utils import create_frames_list_from_traceback

    if use_unicode:

        def method():
            d = {"x": {"á": {"í": {"theta": 1}}}}

            result = d["x"]["á"]["í"]["beta"]

    else:

        def method():
            d = {"x": {"y": {"i": {"theta": 1}}}}

            result = d["x"]["y"]["i"]["beta"]

    try:
        method()
    except:
        exc_type, exc_desc, trace_obj = sys.exc_info()
        memo = {}
        frame = None
        frames_list = create_frames_list_from_traceback(trace_obj, frame, exc_type, exc_desc, memo)
        iter_in = iter(frames_list)
        f = next(iter_in)
        assert f.f_code.co_name == "method"
        line_col_info = frames_list.frame_id_to_line_col_info[id(f)]

        if use_unicode:
            line = '            result = d["x"]["á"]["í"]["beta"]'
        else:
            line = '            result = d["x"]["y"]["i"]["beta"]'

        # Ok, so, the range that we we have covers >>d["x"]["á"]["í"]["beta"]<<
        # the problem here is that ideally we'd like to present to the user that
        # the current key is "beta", so, we need to do some additional computation
        # to find out the proper column to show to the user.
        # (see https://github.com/microsoft/debugpy/issues/1099
        # for more information).
        assert line_col_info.colno == line.index('d["x"]')

        # It's +1 here due to the í unicode char (we need to convert from the bytes
        # index to the actual character in the string to get the actual col).
        if use_unicode:
            assert line_col_info.end_colno == len(line) + 2
        else:
            assert line_col_info.end_colno == len(line)
        original_line = line

        col, endcol = line_col_info.map_columns_to_line(original_line)
        assert col == line.index('["beta"]')
        assert endcol == len(line)


@pytest.mark.parametrize("use_unicode", _USE_UNICODE)
@pytest.mark.skipif(not IS_PY311_OR_GREATER, reason="Python 3.11 required.")
def test_collect_anchors_binop_1(use_unicode):
    from _pydevd_bundle.pydevd_frame_utils import create_frames_list_from_traceback

    if use_unicode:

        def method():
            á = 1
            í = 2
            c = tuple

            result = á + í + c

    else:

        def method():
            a = 1
            b = 2
            c = tuple

            result = a + b + c

    try:
        method()
    except:
        exc_type, exc_desc, trace_obj = sys.exc_info()
        memo = {}
        frame = None
        frames_list = create_frames_list_from_traceback(trace_obj, frame, exc_type, exc_desc, memo)
        iter_in = iter(frames_list)
        f = next(iter_in)
        assert f.f_code.co_name == "method"
        line_col_info = frames_list.frame_id_to_line_col_info[id(f)]

        if use_unicode:
            line = "            result = á + í + c"
            expected_index = line.index("á + í")
        else:
            line = "            result = a + b + c"
            expected_index = line.index("a + b")

        assert line_col_info.colno == expected_index

        # It's +2 here due to the á and í unicode chars (we need to convert from the bytes
        # index to the actual character in the string to get the actual col).
        if use_unicode:
            assert line_col_info.end_colno == len(line) + 2
        else:
            assert line_col_info.end_colno == len(line)
        original_line = line

        col, endcol = line_col_info.map_columns_to_line(original_line)
        assert col == line.index("+ c")
        assert endcol == col + 1


@pytest.mark.parametrize("use_unicode", _USE_UNICODE)
@pytest.mark.skipif(not IS_PY311_OR_GREATER, reason="Python 3.11 required.")
def test_collect_anchors_binop_2(use_unicode):
    from _pydevd_bundle.pydevd_frame_utils import create_frames_list_from_traceback

    if use_unicode:

        def method():
            á = 1
            í = 2
            c = tuple

            result = á + c + í

    else:

        def method():
            a = 1
            b = 2
            c = tuple

            result = a + c + b

    try:
        method()
    except:
        exc_type, exc_desc, trace_obj = sys.exc_info()
        memo = {}
        frame = None
        frames_list = create_frames_list_from_traceback(trace_obj, frame, exc_type, exc_desc, memo)
        iter_in = iter(frames_list)
        f = next(iter_in)
        assert f.f_code.co_name == "method"
        line_col_info = frames_list.frame_id_to_line_col_info[id(f)]

        if use_unicode:
            line = "            result = á + c + í"
            expected_index = line.index("á + c")
        else:
            line = "            result = a + c + b"
            expected_index = line.index("a + c")

        assert line_col_info.colno == expected_index

        # It's +2 here due to the á and í unicode chars (we need to convert from the bytes
        # index to the actual character in the string to get the actual col).
        if use_unicode:
            assert line_col_info.end_colno == line.index("c + í") + 2
        else:
            assert line_col_info.end_colno == line.index("c + b") + 1
        original_line = line

        col, endcol = line_col_info.map_columns_to_line(original_line)
        assert col == 23
        assert endcol == 24
        assert col == line.index("+ c")
        assert endcol == col + 1



#!/usr/bin/env python
"""Simple pyglet example to manually test event loop integration.

To run this:
1) Enable the PyDev GUI event loop integration for pyglet
2) do an execfile on this script
3) ensure you have a working GUI simultaneously with an
   interactive console
"""

if __name__ == "__main__":
    import pyglet

    window = pyglet.window.Window()
    label = pyglet.text.Label(
        "Hello, world",
        font_name="Times New Roman",
        font_size=36,
        x=window.width // 2,
        y=window.height // 2,
        anchor_x="center",
        anchor_y="center",
    )

    @window.event
    def on_close():
        window.close()

    @window.event
    def on_draw():
        window.clear()
        label.draw()



import sys
import os
from _pydev_bundle import pydev_monkey

sys.path.insert(0, os.path.split(os.path.split(__file__)[0])[0])

import unittest

try:
    import thread
except:
    import _thread as thread  # @UnresolvedImport


# =======================================================================================================================
# TestCase
# =======================================================================================================================
class TestCase(unittest.TestCase):
    def test_start_new_thread(self):
        pydev_monkey.patch_thread_modules()
        try:
            found = {}

            def function(a, b, *args, **kwargs):
                found["a"] = a
                found["b"] = b
                found["args"] = args
                found["kwargs"] = kwargs

            thread.start_new_thread(function, (1, 2, 3, 4), {"d": 1, "e": 2})
            import time

            for _i in range(20):
                if len(found) == 4:
                    break
                time.sleep(0.1)
            else:
                raise AssertionError("Could not get to condition before 2 seconds")

            self.assertEqual({"a": 1, "b": 2, "args": (3, 4), "kwargs": {"e": 2, "d": 1}}, found)
        finally:
            pydev_monkey.undo_patch_thread_modules()

    def test_start_new_thread2(self):
        pydev_monkey.patch_thread_modules()
        try:
            found = {}

            class F(object):
                start_new_thread = thread.start_new_thread

                def start_it(self):
                    try:
                        self.start_new_thread(self.function, (1, 2, 3, 4), {"d": 1, "e": 2})
                    except:
                        import traceback

                        traceback.print_exc()

                def function(self, a, b, *args, **kwargs):
                    found["a"] = a
                    found["b"] = b
                    found["args"] = args
                    found["kwargs"] = kwargs

            f = F()
            f.start_it()
            import time

            for _i in range(20):
                if len(found) == 4:
                    break
                time.sleep(0.1)
            else:
                raise AssertionError("Could not get to condition before 2 seconds")

            self.assertEqual({"a": 1, "b": 2, "args": (3, 4), "kwargs": {"e": 2, "d": 1}}, found)
        finally:
            pydev_monkey.undo_patch_thread_modules()


# =======================================================================================================================
# main
# =======================================================================================================================
if __name__ == "__main__":
    unittest.main()
